/*:
 * @plugindesc あたり判定なしのフォロアを表示します。
 * @author liply
 *
 * @help あたり判定無のフォロアを表示します。
 *
 * Follow Pixel
 * ついていく距離です
 *
 * Behind Pixel
 * 後ろにくっつく距離です。
 *
 * Behind Speed Scale
 * 後ろにくっつく際に移動する速度に掛け算します。
 *
 * Behind Switch
 * このスイッチがオンの場合、後ろにくっつきます。
 *
 * プラグインコマンド：
 * ShowFollower on/off
 * GatherFollower
 *
 * @param Follow Pixel
 * @default 48
 *
 * @param Behind Pixel
 * @default 24
 *
 * @param Behind Switch
 * @default 2,18
 *
 * @param Behind Speed Scale
 * @default 2.0
 *
 */

(function() {
    'use strict';

    var parameters = PluginManager.parameters('liply_Follower');
    var followDistance = +parameters['Follow Pixel']/48 || 1;
    var behindDistance = +parameters['Behind Pixel']/48 || 0.5;
    var behindSwitch = parameters['Behind Switch'].split(',').map(function(v){return +v || 2}) || [2, 18];
    var behindSpeedScale = +parameters['Behind Speed Scale'] || 2.0;

    var gatherFlag = false;

    var _Game_Interpreter_pluginCommand = Game_Interpreter.prototype.pluginCommand;
    Game_Interpreter.prototype.pluginCommand = function(cmd, args){
        switch(cmd.toLowerCase()){
            case 'showfollower':
                switch(args[0]){
                    case 'on':
                        followerSwitch(true);
                        break;
                    case 'off':
                        followerSwitch(false);
                        break;
                }

                break;

            case 'gatherfollower':
                gatherFlag = true;

        }

        return _Game_Interpreter_pluginCommand.apply(this, arguments);
    };


    function followerSwitch(value){
        var key = '_liply_followerSwitch';
        if(value !== undefined){
            $gameVariables[key] = value;
        }else{
            if($gameVariables[key] === undefined){
                $gameVariables[key] = true;
            }

            return $gameVariables[key];
        }
    }

    function getFollower(){
        var key = '_liply_follower';
        $gameVariables[key] = $gameVariables[key] || new Game_Character();

        return $gameVariables[key];
    }

    function calcDistance(x1, y1, x2, y2){
        var dx = x1 - x2;
        var dy = y1 - y2;
        return Math.sqrt(dx*dx + dy*dy);
    }

    function lengthSq(v){
        return v.x*v.x+v.y*v.y;
    }

    function calcDistanceInPixel(from ,to){
        return calcDistance(
            from._realX * $gameMap.tileWidth(),
            from._realY * $gameMap.tileHeight(),
            to._realX * $gameMap.tileWidth(),
            to._realY * $gameMap.tileHeight());
    }

    function calcDistanceInReal(from ,to){
        return calcDistance(
            from._realX ,
            from._realY ,
            to._realX ,
            to._realY );
    }

    function limitedToward(limit, delta, fromX, fromY, toX, toY){
        var dx = toX - fromX;
        var dy = toY - fromY;
        var d = Math.sqrt(dx*dx+dy*dy);
        if(d > 0) {
            var len;
            if (d > limit + delta) {
                len = delta / d;
            } else {
                len = (d - limit) / d;
            }

            return {x: dx * len, y: dy * len};
        }else{
            return {x:0 ,y: 0};
        }
    }

    function vectorToward(delta, fromX, fromY, toX, toY){
        var dx = toX - fromX;
        var dy = toY - fromY;
        var d = Math.sqrt(dx*dx+dy*dy);
        if(d > 0) {
            var len;
            if (d > delta) {
                len = delta / d;
            } else {
                len = 1;
            }

            return {x: dx * len, y: dy * len};
        }else{
            return {x:0 ,y: 0};
        }
    }

    function moveTowardCharacter(character, target, limit, delta){
        var distanceInReal = calcDistanceInReal(character, target);
        var real;

        if(limit === 0){
            real = vectorToward(delta,
                character._realX, character._realY,
                target._realX, target._realY - 2/48);

            character._realX += real.x;
            character._realY += real.y;

            character._liply_moveFlag = true;
            if(lengthSq(real) < (2/48)*(2/48))
                return false;
            else
                return true;
        }else if(distanceInReal > limit + 0.01){
            real =
                limitedToward(limit, delta, character._realX, character._realY, target._realX, target._realY);

            character._realX += real.x;
            character._realY += real.y;
            character._liply_moveFlag = true;

            return true;
        }else{
            character._liply_moveFlag = false;

            return false;
        }
    }

    function moveBehindCharacter(character, target, limit, delta){
        var targetX = 0, targetY = 0;
        var sq2 = Math.sqrt(2);

        switch(target.direction()){
            case 2:
                targetX = behindDistance / sq2;
                targetY = -behindDistance / sq2;
                break;

            case 4:
                targetX = behindDistance / sq2;
                targetY = behindDistance / sq2;
                break;

            case 6:
                targetX = -behindDistance / sq2;
                targetY = behindDistance / sq2;
                break;

            case 8:
                targetX = -behindDistance / sq2;
                targetY = behindDistance / sq2;
                break;
        }
        var real = vectorToward(delta,
            character._realX, character._realY,
            target._realX + targetX, target._realY + targetY);

        if((lengthSq(real)) > 0.00001){
            character._realX += real.x;
            character._realY += real.y;

            character._liply_moveFlag = true;
        }else{
            character._liply_moveFlag = false;
        }
    }

    function isWarp(character, target){
        return calcDistanceInPixel(character, target) > $gameMap.tileWidth()* 2;
    }

    function updateFollower(follower){
        if($gameParty.battleMembers().length >= 2){
            var actor = $gameParty.battleMembers()[1];
            follower.setImage(actor.characterName(), actor.characterIndex());

            if (follower._liply_moveFlag && follower.hasWalkAnime()) {
                follower._animationCount += 1.5;
            } else if (follower.hasStepAnime() || !follower.isOriginalPattern()) {
                follower._animationCount++;
            }

            if (follower._animationCount >= follower.animationWait()) {
                follower.updatePattern();
                follower._animationCount = 0;
            }

            return true;
        }

        return false;
    }

    function aimCharacter(from, to){
        var sx = to._realX - from._realX;
        var sy = to._realY - from._realY;
        if (Math.abs(sx) > Math.abs(sy)) {
            from.setDirection(sx < 0 ? 4 : 6);
        } else if (sy !== 0) {
            from.setDirection(sy < 0 ? 8 : 2);
        }
    }
    var spriteFollower;
    var spritePlayer;
    var Spriteset_Map_createCharacters = Spriteset_Map.prototype.createCharacters;
    Spriteset_Map.prototype.createCharacters = function() {
        Spriteset_Map_createCharacters.apply(this, arguments);
        spritePlayer = this._characterSprites[this._characterSprites.length-1];

        spriteFollower = new Sprite_Character(getFollower());
        this._characterSprites.push(spriteFollower);
        this._tilemap.addChild(spriteFollower);
    };

    function isBehindSwitchOn(switches){
        for(var n = 0; n < switches.length; ++n){
            if($gameSwitches.value(switches[n])) return true;
        }

        return false;
    }

    var Game_Player_update = Game_Player.prototype.update;
    Game_Player.prototype.update = function(){
        Game_Player_update.apply(this, arguments);
        var follower = getFollower();

        if(followerSwitch()){

            if(!updateFollower(follower)){
                follower.setTransparent(true);
            }else{
                if(isWarp(follower, $gamePlayer)){
                    follower._realX = $gamePlayer._realX;
                    follower._realY = $gamePlayer._realY - 0.2;
                }

                if(gatherFlag){
                    if(!moveTowardCharacter(follower, $gamePlayer, 0, $gamePlayer.distancePerFrame())){
                        gatherFlag = false;
                    }
                    aimCharacter(follower, $gamePlayer);

                }else if(isBehindSwitchOn(behindSwitch)){
                    moveBehindCharacter(follower, $gamePlayer, behindDistance,
                        $gamePlayer.distancePerFrame() * behindSpeedScale);
                    follower.setDirection($gamePlayer.direction());

                }else{
                    moveTowardCharacter(follower, $gamePlayer, followDistance, $gamePlayer.distancePerFrame());
                    aimCharacter(follower, $gamePlayer);
                }

                follower.setTransparent(false);
                spriteFollower.blendMode = spritePlayer.blendMode;
                spriteFollower.setBlendColor(spritePlayer.getBlendColor());
            }
        }else{
            follower.setTransparent(true);
        }
    };

})();